/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation. See the file COPYING
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Copyright (C) 2000-2007 Frank van Maarseveen <frankvm@frankvm.com>
 */

#include "fsindex.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>

#include "md5.h"
#include "sha256.h"
#include "repr.h"
#include "digest.h"

unsigned long long bytes_digested;
static void xclose(int d, const char *path);

void digest_none(struct inode *inode)
{
}

void digest_file_size(struct inode *inode)
{
	char buf[100];

	sprintf(buf, "%llx", inode->size);
	inode->digest = alcstr(buf);
}

void digest_file_read(struct inode *inode)
{
	int d, n;
	char buf[65536];
	const char *path;

	path = fs_path(inode->dentry);
	d = fs_open(path, 0);
	if (d < 0)
		return;
	do {
		n = read(d, buf, sizeof (buf));
	} while (n > 0);
	if (n < 0)
		warn("read %s: %s\n", path, strerror(errno));
	xclose(d, path);
	digest_file_size(inode);
}

void digest_file_md5(struct inode *inode)
{
	unsigned char md5sum[16];
	int d;
	const char *path;

	path = fs_path(inode->dentry);
	d = fs_open(path, 0);
	if (d < 0)
		return;
	if (md5_stream(d, md5sum))
		warn("read %s: %s\n", path, strerror(errno));
	else
		inode->digest = encode(md5sum, sizeof (md5sum));
	xclose(d, path);
}

void digest_file_sha256(struct inode *inode)
{
	unsigned char sha256sum[32];
	int d;
	const char *path;

	path = fs_path(inode->dentry);
	d = fs_open(path, 0);
	if (d < 0)
		return;
	if (sha256_stream(d, sha256sum))
		warn("read %s: %s\n", path, strerror(errno));
	else
		inode->digest = encode(sha256sum, sizeof (sha256sum));
	xclose(d, path);
}

void digest_symlink(struct inode *inode)
{
	char *buf;
	int n, len;
	const char *path;

	path = fs_path(inode->dentry);
	n = (int)inode->size;
	buf = NULL;
	while (1) {
		buf = xrealloc(buf, n + 1);
		len = readlink(path, buf, n + 1);
		if (len < 0) {
			warn("readlink %s: %s\n", path, strerror(errno));
			free(buf);
			return;
		}
		if (len < n)
			buf = xrealloc(buf, len + 1);
		if (len <= n) {
			buf[len] = 0;
			break;
		}
		n = 2 * n + 1;
	}
	inode->digest = escape(buf);
	if (inode->digest != buf)
		free(buf);
}

void digest_device(struct inode *inode)
{
	char buf[100];

	sprintf(buf, "%llx", (unsigned long long)inode->rdev);
	inode->digest = alcstr(buf);
}

static void xclose(int d, const char *path)
{
	off_t pos;

	pos = lseek(d, 0LL, SEEK_CUR);
	if (pos == (off_t)-1)
		die("lseek %s: %s\n", path, strerror(errno));
	else
		bytes_digested += pos;
	if (close(d))
		die("close %s: %s\n", path, strerror(errno));
}
