/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation. See the file COPYING
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Copyright (C) 2000-2007 Frank van Maarseveen <frankvm@frankvm.com>
 */

#include "fsindex.h"

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <dirent.h>
#define __USE_GNU 1		/* for O_NOATIME and O_DIRECTORY */
#include <fcntl.h>
#include <errno.h>
#include <sys/syscall.h>	/* __NR_getdents64: need it for O_NOATIME on directories */

#if __linux__ && !O_NOATIME
# if __i386__
#  warning "guessing O_NOATIME!!!"
#  define O_NOATIME 01000000
   /* ugly as hell but at least test the dangerous and nearby constants */
#  if O_NOATIME & (O_CREAT|O_TRUNC|O_WRONLY|O_RDWR|O_DIRECTORY|O_NOFOLLOW)
#   error "bad O_NOATIME"
#  endif
# endif
#endif

#include "fs.h"

#define ISROOT(d)		(!(d)->parent)

static struct dentry *		root;
static struct radix_tree *	devices;
static bool_t			more_mounts;
static void			fs_fill(void);
static void			fs_fill_inode(struct device *device, struct dentry *head);
static void			fs_fill_dir(struct dentry *dir, const char *path, struct device *device);
static struct inode *		fs_new_inode(struct radix_tree *irt, struct stat *st);
static void			fs_del_inode(struct inode *inode);
static struct device *		fs_new_device(dev_t dev);
static void			fs_del_device(struct device *device);
static void			fs_add_dirent(struct dentry *dir, const struct dirent *dirent, struct device *device);
static void			fs_add_dentry(struct dentry *dir, const char *name, ino_t ino, struct device *device);
static void			fs_del_dentry(struct dentry *dentry);
static int			fs_cmp_dentry(const void *v1, const void *v2);
static void			fs_treewalk(struct dentry *dir, void (*f)(struct dentry *dentry));
static void			fs_dumpdentry(void *data);
static void			fs_dumpinode(void *data);

static unsigned long long kilobytes;
static unsigned long long dentries;
static unsigned long long inodes;
static unsigned long long directories;
static int maxpath;

struct fsinfo fs_init(int argc, char **argv)
{
	struct stat st;
	struct fsinfo fsinfo = {0};

	int i;

	BUG_ON(sizeof (ino_t) > sizeof (unsigned long long));
	BUG_ON(sizeof (dev_t) > sizeof (unsigned long long));
	devices = rt_new();
	root = alc(sizeof (*root));
	for (i = 0; i < argc; ++i) {
		if (lstat(argv[i], &st) == -1) {
			warn("%s: %s\n", argv[i], strerror(errno));
			continue;
		}
		fs_add_dentry(root, argv[i], st.st_ino, fs_new_device(st.st_dev));
	}
	fs_fill();
	fsinfo.kilobytes = kilobytes;
	fsinfo.dentries = dentries;
	fsinfo.inodes = inodes;
	fsinfo.directories = directories;
	fsinfo.maxpath = maxpath;
	return fsinfo;
}

/*
 * Return the full pathname. The return value points to static
 * storage which will be reused on the next call.
 */
static char *fs_path_buffer;
const char *fs_path(struct dentry *dentry)
{
	static char *end;
	char *ptr, *p, *name, sep;
	int len, used, newsize;

	ptr = end;
	sep = 0;
	do {
		name = dentry->name;
		len = strlen(name);
		if (len && name[len - 1] == '/' && sep)
			--len;		/* trim trailing slash from fsindex argument */
		if (ptr - fs_path_buffer < len + 1) {
			used = end - ptr;
			newsize = 2 * used + len + 1;
			p = xmalloc(newsize);
			end = p + newsize;
			memcpy(end - used, ptr, used);
			free(fs_path_buffer);
			fs_path_buffer = p;
			ptr = end - used;
		}
		*--ptr = sep;
		sep = '/';
		ptr -= len;
		memcpy(ptr, name, len);
		dentry = dentry->parent;
	} while (!ISROOT(dentry));
	if (end - ptr - 1> maxpath)
		maxpath = end - ptr - 1;
	return ptr;
}

int fs_open(const char *path, int dirflags)
{
	int d, flags;

	flags = O_RDONLY | dirflags;
#if __linux__
	if (opt_noatime)
		flags |= O_NOATIME;
#endif
	d = open(path, flags);
	if (d == -1)
		warn("cannot open%s %s: %s\n", dirflags ? " directory" : "", path, strerror(errno));
#if 0
	else if (!dirflags) {
		syscall(__NR_fadvice64_64, d, (off_t)0, (off_t)0, FADV_SEQUENTIAL);
		syscall(__NR_fadvice64_64, d, (off_t)0, (off_t)0, FADV_NOREUSE);
	}
#endif
	return d;
}

void fs_walk_inodes(void (*f)(struct inode *inode))
{
	void *buf[1024];
	unsigned long long nextdev, nextino;
	struct device *device;
	int i, n;

	nextdev = 0;
	do {
		if (!rt_read(devices, (void **)&device, 1, &nextdev))
			break;
		nextino = 0;
		do {
			n = rt_read(device->inode_rt, buf, ARRAYSIZE(buf), &nextino);
			for (i = 0; i < n; ++i)
				f(buf[i]);
		} while (nextino);
	} while (nextdev);
}

/*
 * Sorted pre-order tree traversal.
 */
void fs_walk_dentries(void (*f)(struct dentry *dentry))
{
	fs_treewalk(root, f);
}

void fs_cleanup(void)
{
	unsigned long long nextdev;
	struct device *device;

	nextdev = 0;
	do {
		if (!rt_read(devices, (void **)&device, 1, &nextdev))
			break;
		fs_del_device(device);
	} while (nextdev);
	rt_del(devices);
	free(root);
	free(fs_path_buffer);
	BUG_ON(inodes);
	BUG_ON(dentries);
}

void fs_dump(void)
{
	unsigned long long nextdev;
	struct device *device;

	rt_dump(devices, NULL, "devices");
	nextdev = 0;
	do {
		if (!rt_read(devices, (void **)&device, 1, &nextdev))
			break;
		rt_dump(device->dentry_rt, fs_dumpdentry, "%llx->dentry_rt", (unsigned long long)device->dev);
		rt_dump(device->inode_rt, fs_dumpinode, "%llx->inode_rt", (unsigned long long)device->dev);
	} while (nextdev);
}

/*
 * Expand and transform radix tree dentry_rt into inode_rt, for every device.
 * This requires lstat() on all pathnames (in inode number order to minimize
 * disk seeks) and adding directory contents to dentry_rt again.
 */
static void fs_fill(void)
{
	void *buf[1024];
	int i, n;
	unsigned long long nextdev, nextino;
	struct device *device;
	struct dentry *dentry;

	do {
		more_mounts = false;
		nextdev = 0;
		do {
			if (!rt_read(devices, (void **)&device, 1, &nextdev))
				break;
			do {
				device->more_dentries = false;
				nextino = 0;
				do {
					n = rt_read(device->dentry_rt, buf, ARRAYSIZE(buf), &nextino);
					for (i = 0; i < n; ++i) {
						dentry = buf[i];
						rt_remove(device->dentry_rt, dentry->ino);
						fs_fill_inode(device, dentry);
					}
				} while (nextino);
			} while (device->more_dentries);
		} while (nextdev);
	} while (more_mounts);
}

static void fs_fill_inode(struct device *device, struct dentry *head)
{
	struct device *mnt_device;
	struct dentry *dentry, **slot;
	struct inode *inode;
	const char *path, **pp;
	struct stat st;

	dentry = head;
	inode = NULL;
	do {
		LIST_NAME_CUT(head, links, dentry);
		path = fs_path(dentry);
		for (pp = opt_exclude; *pp; ++pp) {
			if (strcmp(*pp, path) == 0)
				break;
		}
		if (*pp) {
			/* -Xpath */
		} else if (lstat(path, &st) == -1) {
			warn("lstat %s: %s\n", path, strerror(errno));
		} else if (st.st_dev != device->dev) {
			for (pp = opt_include; *pp; ++pp) {
				if (strcmp(*pp, path) == 0)
					break;
			}
			if (!opt_onefs || *pp) {
				/* maybe dirent->d_ino disagrees with st.st_ino for mount points so fix it */
				dentry->ino = st.st_ino;
				mnt_device = fs_new_device(st.st_dev);
				slot = (struct dentry **)rt_insert_slot(mnt_device->dentry_rt, st.st_ino);
				LIST_NAME_ADD(*slot, links, dentry);
				more_mounts = true;
				continue;
			}
		} else if (st.st_ino != dentry->ino) {
			warn("%s: file has been replaced, skipping.\n", path);
		} else {
			if (!inode)
				inode = fs_new_inode(device->inode_rt, &st);
			LIST_NAME_ADD(inode->dentry, links, dentry);
			dentry->inode = inode;
			++inode->nlink;
			if (S_ISDIR(st.st_mode)) {
				fs_fill_dir(dentry, path, device);
				++directories;
			}
			continue;
		}
		/* pathname didn't work out */
		LIST_NAME_CUT(dentry->parent->childs, childs, dentry);
		--dentry->parent->nchild;
		fs_del_dentry(dentry);
	} while ((dentry = head));
}

static void fs_fill_dir(struct dentry *dir, const char *path, struct device *device)
{
	struct dirent *dirent;
#if __linux__
	char buf[0x10000];
	int d, n;

	d = fs_open(path, O_DIRECTORY);
	if (d == -1)
		return;
	while ((n = syscall(__NR_getdents64, d, buf, sizeof (buf))) > 0) {
		dirent = (struct dirent *)buf;
		do {
			fs_add_dirent(dir, dirent, device);
			n -= dirent->d_reclen;
			dirent = (struct dirent *)((char *)dirent + dirent->d_reclen);
		} while (n > 0);
	}
	if (n < 0)
		die("getdents64 %s: %s\n", path, strerror(errno));
	if (close(d))
		die("close directory %s: %s\n", path, strerror(errno));
#else
	DIR *d;

	d = opendir(path);
	if (!d) {
		warn("opendir %s: %s\n", path, strerror(errno));
		return;
	}
	while ((dirent = readdir(d)))
		fs_add_dirent(dir, dirent, device);
	if (closedir(d))
		die("closedir %s: %s\n", path, strerror(errno));
#endif
}

static struct inode *fs_new_inode(struct radix_tree *irt, struct stat *st)
{
	struct inode *inode, **slot;

	slot = (struct inode **)rt_insert_slot(irt, st->st_ino);
	if (*slot) {
		inode = *slot;
	} else {
		inode = alc(sizeof (struct inode));
		inode->ino = st->st_ino;
		inode->mode = st->st_mode;
		inode->uid = st->st_uid;
		inode->gid = st->st_gid;
		inode->rdev = st->st_rdev;
		inode->size = st->st_size;
		inode->mtime = st->st_mtime;
		*slot = inode;
		++inodes;
		if (S_ISREG(st->st_mode))
			kilobytes += ((st->st_size + 1023) >> 10);
	}
	return inode;
}

static void fs_del_inode(struct inode *inode)
{
	free(inode->digest);
	free(inode);
	--inodes;
}

static struct device *fs_new_device(dev_t dev)
{
	struct device *device, **slot;

	slot = (struct device **)rt_insert_slot(devices, dev);
	device = *slot;
	if (!device) {
		*slot = device = alc(sizeof (*device));
		device->dev = dev;
		device->dentry_rt = rt_new();
		device->inode_rt = rt_new();
	}
	return device;
}

static void fs_del_device(struct device *device)
{
	void *buf[1024];
	int i, n;
	unsigned long long nextino;
	struct inode *inode;
	struct dentry *dentry, *head;

	rt_del(device->dentry_rt);
	nextino = 0;
	do {
		n = rt_read(device->inode_rt, buf, ARRAYSIZE(buf), &nextino);
		for (i = 0; i < n; ++i) {
			inode = buf[i];
			dentry = head = inode->dentry;
			do {
				LIST_NAME_CUT(head, links, dentry);
				fs_del_dentry(dentry);
			} while ((dentry = head));
			fs_del_inode(inode);
		}
	} while (nextino);
	rt_del(device->inode_rt);
	free(device);
}

static void fs_add_dirent(struct dentry *dir, const struct dirent *dirent, struct device *device)
{
	const char *name;

	name = dirent->d_name;
	if (strcmp(name, ".") && strcmp(name, "..")) {
		fs_add_dentry(dir, name, dirent->d_ino, device);
	}
}

static void fs_add_dentry(struct dentry *dir, const char *name, ino_t ino, struct device *device)
{
	struct dentry *dentry, **slot;

	dentry = alc(sizeof (*dentry));
	dentry->parent = dir;
	LIST_NAME_ADD(dir->childs, childs, dentry);
	++dir->nchild;
	dentry->name = alcstr(name);
	dentry->ino = ino;
	slot = (struct dentry **)rt_insert_slot(device->dentry_rt, ino);
	LIST_NAME_ADD(*slot, links, dentry);
	++dentries;
	device->more_dentries = true;
}

static void fs_del_dentry(struct dentry *dentry)
{
	free(dentry->name);
	free(dentry);
	--dentries;
}

static int fs_cmp_dentry(const void *v1, const void *v2)
{
	struct dentry *const *p1 = v1;
	struct dentry *const *p2 = v2;

	return strcmp((*p1)->name, (*p2)->name);
}

static void fs_treewalk(struct dentry *dir, void (*f)(struct dentry *dentry))
{
	struct dentry **buf, *head, *dentry;
	int i, len;

	len = dir->nchild;
	buf = xmalloc(len * sizeof (*buf));
	dentry = head = dir->childs;
	for (i = 0; i < len; ++i) {
		BUG_ON(!dentry);
		buf[i] = dentry;
		dentry = LIST_NAME_NEXT(head, childs, dentry);
	}
	BUG_ON(dentry);
	qsort(buf, len, sizeof (*buf), fs_cmp_dentry);
	if (!ISROOT(dir))
		f(dir);
	for (i = 0; i < len; ++i) {
		dentry = buf[i];
		if (dentry->childs)
			fs_treewalk(dentry, f);
		else
			f(dentry);
	}
	free(buf);
}

static void fs_dumpdentry(void *data)
{
	struct dentry *head, *p;

	head = data;
	printf("%llu", head->ino);
	for (p = head; p; p = LIST_NAME_NEXT(head, links, p))
		printf(" %s", p->name);
	printf("\n");
}

static void fs_dumpinode(void *data)
{
	printf("%llu\n", ((struct inode *)data)->ino);
}
