/*
 * fsindex, create a file system index in a kind of "ls -l" format, including
 * MD5 sums.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation. See the file COPYING
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Copyright (C) 2000-2007 Frank van Maarseveen <frankvm@frankvm.com>
 */

#include "fsindex.h"

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <time.h>

#include "digest.h"
#include "repr.h"
#include "fs.h"

#define ME "fsindex"

static const char *	usage = 
"Usage: " ME " [options...] [arg...] [options...]\n"
"\n"
"Option summary (see \"man " ME "\" for more details):\n"
"\n"
"  -d<digest>	choose digest method for regular files (default md5):\n"
"\n"
"		none	no digest.\n"
"		size	hexadecimal size in bytes according to stat(2).\n"
"		read	like size, but do read the file (media check).\n"
"		md5	base 64 MD5.\n"
"		sha256	base 64 SHA256 (FIPS-180-2).\n"
"\n"
"  -i		print inode number. This adds a space plus inode number in hex.\n"
"  -I<path>	include <path> in the traversal, even when it's a mount point\n"
"		and -x has been specified. This option accumulates.\n"
"  -l		print link count. This adds a space plus in-tree link count.\n"
"  -n		print numerical user and group ids instead of name.\n"
"  -v		print time stamps and some statistics on standard error.\n"
"  -x		do not cross mount points.\n"
"  -X<path>	exclude <path> from the traversal. This option accumulates.\n"
"  --debug	print debug information (currently: radix trees) before exit.\n"
"  --hex\t	print md5 and sha256 digests in hexadecimal notation.\n"
"  --localtime	use the local time zone instead of UTC.\n"
#if __linux__
"  --noatime	preserve access times. This option is always on when the\n"
"		effective user is root.\n"
#endif
;

struct digest {
	const char *	name;
	void		(*func)(struct inode *);
};

static const struct digest digest_tab[] = {
	{ "none",	digest_none		},
	{ "read",	digest_file_read	},
	{ "size",	digest_file_size	},
	{ "md5",	digest_file_md5		},
	{ "sha256",	digest_file_sha256	},
	{ NULL,		NULL			}
};

static int		exit_status;
static char *		nextarg(int argc, char *argv[], int *pi, char **pp);
static void		digest(struct inode *inode);
static void		print_signature(struct dentry *dentry);
static void		print_vmpeak(void);
static void		(*opt_digest)(struct inode *) = digest_file_md5;
static bool_t		opt_inode;
const char **		opt_include;
static bool_t		opt_nlink;
bool_t			opt_numeric;
bool_t			opt_onefs;
static int		opt_verbose;
static FILE *		verbose_out;
const char **		opt_exclude;
bool_t			opt_debug;
bool_t			opt_hex;
static bool_t		opt_localtime;
bool_t			opt_noatime;

int main(int argc, char **argv)
{
	int i, ac, ic, xc;
	bool_t optend;
	char **av, *p, *a;
	const struct digest *dg;
	struct fsinfo fsinfo;

	ac = 0;
	av = alc((1 + argc) * sizeof (char *));		/* +1: there's a default added */
	ic = 0;
	opt_include = alc(argc * sizeof (char *));
	xc = 0;
	opt_exclude = alc(argc * sizeof (char *));
	opt_noatime = geteuid() == 0;
	optend = false;;
	for (i = 1; i < argc; ++i) {
		if (optend || argv[i][0] != '-') {
			av[ac++] = argv[i];
			continue;
		}
		if (strcmp(argv[i], "--debug") == 0) {
			opt_debug = true;
			continue;
		}
		if (strcmp(argv[i], "--hex") == 0) {
			opt_hex = true;
			continue;
		}
		if (strcmp(argv[i], "--localtime") == 0) {
			opt_localtime = true;
			continue;
		}
#if __linux__
		if (strcmp(argv[i], "--noatime") == 0) {
			opt_noatime = true;
			continue;
		}
#endif
		if (argv[i][1] == '-') {
			if (argv[i][2])
				die("%s: unknown option. Type \"" ME " -?\" for help\n", argv[i]);
			optend = true;
			continue;
		}
		for (p = argv[i] + 1; *p; ++p) {
			switch (*p) {
			case 'd':
				a = nextarg(argc, argv, &i, &p);
				for (dg = digest_tab; dg->name; ++dg) {
					if (strcmp(dg->name, a) == 0) {
						opt_digest = dg->func;
						break;
					}
				}
				if (!dg->name)
					die("-d %s: unknown digest method.\n", a);
				break;
			case 'i':
				opt_inode = true;
				break;
			case 'I':
				opt_include[ic++] = nextarg(argc, argv, &i, &p);
				break;
			case 'l':
				opt_nlink = true;
				break;
			case 'n':
				opt_numeric = true;
				break;
			case 'v':
				++opt_verbose;
				break;
			case 'x':
				opt_onefs = true;
				break;
			case 'X':
				opt_exclude[xc++] = nextarg(argc, argv, &i, &p);
				break;
			case '?':
				puts(usage);
				exit(0);
			default:
				die("-%c: unknown option. Type \"" ME " -?\" for help\n", *p);
			}
		}
	}
	if (!ac) {
		av[ac++] = ".";
	}
	av[ac] = NULL;
	opt_include[ic] = NULL;
	opt_exclude[xc] = NULL;
	verbose_out = opt_verbose ? stderr : fopen("/dev/null", "w");
	if (!verbose_out)
		die("opening /dev/null: %s\n", strerror(errno));
	repr_init();
	fprintf(verbose_out, "%s indexing...\n", time2str(time(NULL), false));
	fsinfo = fs_init(ac, av);
	fprintf(verbose_out, "%llu kB data, %llu dentries, %llu inodes, %llu directories, maxpath=%d\n",
		fsinfo.kilobytes, fsinfo.dentries, fsinfo.inodes, fsinfo.directories, fsinfo.maxpath);
	fprintf(verbose_out, "%s digesting...\n", time2str(time(NULL), false));
	fs_walk_inodes(digest);
	fprintf(verbose_out, "%s writing...\n", time2str(time(NULL), false));
	fs_walk_dentries(print_signature);
	fprintf(verbose_out, "%s done\n", time2str(time(NULL), false));
	if (opt_debug)
		fs_dump();
	repr_cleanup();
	fs_cleanup();
	free(av);
	free(opt_include);
	free(opt_exclude);
	fflush(stdout);
	if (ferror(stdout))
		die("standard output: %s\n", strerror(errno));
	print_vmpeak();
	if (verbose_out != stderr)
		fclose(verbose_out);
	return exit_status;
}

static char *nextarg(int argc, char *argv[], int *pi, char **pp)
{
        char    option;
        char    *arg;

        option = **pp;
        arg = *pp + 1;
        if (*arg)
                *pp = " ";
        else if (++*pi < argc)
                arg = argv[*pi];
        else
                die("-%c: missing argument.\n", option);
        return arg;
}

static void digest(struct inode *inode)
{
	switch (inode->mode & S_IFMT) {
	case S_IFREG:
		opt_digest(inode);
		break;
	case S_IFLNK:
		digest_symlink(inode);
		break;
	case S_IFBLK:
	case S_IFCHR:
		digest_device(inode);
		break;
	case S_IFDIR:
	case S_IFSOCK:
	case S_IFIFO:
#ifdef S_IFDOOR
	case S_IFDOOR:		/* Solaris */
#endif
		break;
	default:
		warn("%s has unknown type 0%lo\n", fs_path(inode->dentry), (unsigned long)(inode->mode & S_IFMT));
	}
}

/*
 * The signature of a file system object is comprised of anything which would
 * be "interesting" enough for synchronizing file-systems or taking back-ups. But
 * in particular the following is not interesting:
 *
 * -	date/time last modified of anything else but a plain file or directory.
 * -	symlink permission bits.
 */
static void print_signature(struct dentry *dentry)
{
	const char *modestr, *timdatstr, *path, *digeststr;
	struct inode *inode;
	char *pathstr;

	inode = dentry->inode;
	modestr = "lrwxr-xr-x";
	if (!S_ISLNK(inode->mode))
		modestr = mode2str(inode->mode);
	timdatstr = "-------- --------";
	if (S_ISDIR(inode->mode) || S_ISREG(inode->mode))
		timdatstr = time2str(inode->mtime, !opt_localtime);
	digeststr = "-";
	if (inode->digest)
		digeststr = inode->digest;
	path = fs_path(dentry);
	pathstr = escape(path);
	printf("%s %s %s\t", modestr, uid2str(inode->uid), gid2str(inode->gid));
	printf("%s %s\t", timdatstr, pathstr);
	printf("%s", digeststr);
	if (opt_inode)
		printf(" %llx", (unsigned long long)inode->ino);
	if (opt_nlink)
		printf(" %d", inode->nlink);
	printf("\n");
	if (path != pathstr)
		free(pathstr);
	if (ferror(stdout))
		die("standard output: %s\n", strerror(errno));
}

static void print_vmpeak(void)
{
	FILE *f;
	char buf[1000];
	unsigned long vmpeak;

	if (!opt_verbose)
		return;
	f = fopen("/proc/self/status", "r");
	if (f) {
		while (fgets(buf, sizeof (buf), f) == buf) {
			if (sscanf(buf, "VmPeak:%lu", &vmpeak) == 1) {
				fprintf(verbose_out, "Memory used: %lu kB\n", vmpeak);
				break;
			}
		}
		fclose(f);
	}
}

void *xmalloc(int size)
{
	void *mem;

	mem = malloc(size);
	if (!mem)
		die("cannot allocate %d bytes.\n", size);
	return mem;
}

void *xrealloc(void *mem, int size)
{
	mem = realloc(mem, size);
	if (size && !mem)
		die("cannot reallocate %d bytes.\n", size);
	return mem;
}

void *alc(int size)
{
	void	*mem;

	mem = xmalloc(size);
	memset(mem, 0, size);
	return mem;
}

char *alcstr(const char *s)
{
	char	*copy;
	int	size;

	size = strlen(s) + 1;
	copy = xmalloc(size);
	memcpy(copy, s, size);
	return copy;
}

void warn(const char *fmt, ...)
{
	va_list	ap;

	fprintf(stderr, ME ": ");
	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	exit_status = 1;
}

void die(const char *fmt, ...)
{
	va_list	ap;

	fprintf(stderr, ME ": ");
	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	exit(2);
}
