/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation. See the file COPYING
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Copyright (C) 2000-2007 Frank van Maarseveen <frankvm@frankvm.com>
 */

#include "fsindex.h"

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <stdarg.h>

#include "rt.h"

#define INDEXBITS		(8 * (sizeof (unsigned long long)))
#define NODEBITS		6
#define NODESLOTS		(1ULL << NODEBITS)
#define NODEMASK		(NODESLOTS - 1)

struct node {
	struct node *		slots[NODESLOTS];
	unsigned int		count;		/* track # slots in use so we can prune the tree */
};

struct radix_tree {
	struct node *		root;
	unsigned int		maxbits;	/* multiple of NODEBITS, possibly >= INDEXBITS */
	unsigned long long	maxindex;	/* The all-ones index with maxbits # of bits */
};

struct slotpath {
	struct node **		slot;
	struct node **		end;
};

struct radix_tree *rt_new(void)
{
	return alc(sizeof (struct radix_tree));
}

/*
 * Free an entire radix tree. Dumb implementation because we can't
 * determine index of item. Radix tree users probably can do better.
 */
void rt_del(struct radix_tree *rt)
{
	unsigned long long index;

	index = 0;
	do {
		rt_remove(rt, index);
		rt_read(rt, NULL, 0, &index);
	} while (index);
	BUG_ON(rt->maxindex || rt->maxbits || rt->root);
	free(rt);
}

/*
 * Return the address of the slot where to store an item with specified index.
 * If the slot doesn't exist it will be created holding NULL. In that case the
 * caller is required to store a nonzero value there afterwards.
 */
void **rt_insert_slot(struct radix_tree *rt, unsigned long long index)
{
	unsigned int bits;
	struct node **slot, *node;
	unsigned int *countp, rootcount = 0;

	/*
	 * Increase tree height so we don't accidentally dereference an item
	 * thinking it's a node. To avoid a redundant chain of nodes for
	 * index==0 ending with a node having a zero count we skip node
	 * allocation when there are no items.
	 */
	while (index > rt->maxindex) {
		if (rt->root) {
			node = alc(sizeof (struct node));
			node->slots[0] = rt->root;
			node->count = 1;
			rt->root = node;
		}
		rt->maxbits += NODEBITS;
		rt->maxindex = (rt->maxindex << NODEBITS) | NODEMASK;
	}
	slot = &rt->root;
	countp = &rootcount;		/* result discarded */
	for (bits = rt->maxbits; bits; bits -= NODEBITS) {
		if (!*slot) {
			*slot = alc(sizeof (struct node));
			++*countp;
		}
		countp = &(*slot)->count;
		slot = (*slot)->slots + ((index >> (bits - NODEBITS)) & NODEMASK);
	}
	if (!*slot)
		++*countp;
	return (void **)slot;
}

/*
 * Remove an item from the tree, returning it. Return NULL if not found.
 */
void *rt_remove(struct radix_tree *rt, unsigned long long index)
{
	void *item;
	unsigned int bits;
	struct node *node, **slot, ***p;
	struct node **path[1 + (INDEXBITS + NODEBITS - 1) / NODEBITS];

	if (index > rt->maxindex)
		return NULL;
	slot = &rt->root;
	path[0] = slot;
	p = path;
	for (bits = rt->maxbits; bits; bits -= NODEBITS) {
		if (!*slot)
			break;
		slot = (*slot)->slots + ((index >> (bits - NODEBITS)) & NODEMASK);
		*++p = slot;
	}
	item = *slot;
	if (item) {
		*slot = NULL;
		for (--p; p >= path; --p) {
			if (--(**p)->count)
				break;
			free(**p);
			**p = NULL;
		}

		/*
		 * Reduce tree height if possible. When there are no items then rt->root will
		 * be NULL: the last chain of nodes has completely been eaten by above loop.
		 */
		if (rt->root) {
			while (rt->maxbits && rt->root->count == 1 && rt->root->slots[0]) {
				rt->maxbits -= NODEBITS;
				rt->maxindex &= ~((unsigned long long)NODEMASK << rt->maxbits);
				node = rt->root;
				rt->root = node->slots[0];
				free(node);
			}
		} else {
			rt->maxbits = 0;
			rt->maxindex = 0;
		}
	}
	return item;
}

/*
 * Return item stored at specified index or NULL if it doesn't exist.
 */
void *rt_lookup(struct radix_tree *rt, unsigned long long index)
{
	unsigned int bits;
	struct node *node;

	if (index > rt->maxindex)
		return NULL;
	node = rt->root;
	for (bits = rt->maxbits; bits; bits -= NODEBITS) {
		if (!node)
			break;
		node = node->slots[(index >> (bits - NODEBITS)) & NODEMASK];
	}
	return node;
}

/*
 * Read at most maxitems items (pointers) starting from the specified index
 * and store them in buf. Afterwards, the index is always positioned to point
 * to the first item to read in the next call unless the end of the tree is
 * reached. In that case the index will wrap to zero and the function will
 * return what has been read so far. There is no special value for marking
 * end of tree so reading until rt_read() returns zero won't work: it will
 * only return zero when the tree is empty.
 */
int rt_read(struct radix_tree *rt, void **buf, int maxitems, unsigned long long *indexp)
{
	unsigned int bits;
	unsigned long long index;
	struct node **slot, **end;
	struct slotpath	*p, path[(INDEXBITS + NODEBITS - 1) / NODEBITS];
	int items;

	index = *indexp;
	if (index > rt->maxindex) {
		*indexp = 0;
		return 0;
	}
	p = path;
	slot = &rt->root;
	end = slot + 1;
	for (bits = rt->maxbits; bits; bits -= NODEBITS) {
		if (!*slot) {
			/* align index with current slot path */
			index &= ~(unsigned long long)NODEMASK << (bits - NODEBITS);
			break;
		}
		p->slot = slot;
		p->end = end;
		++p;
		end = (*slot)->slots + NODESLOTS;
		slot = (*slot)->slots + ((index >> (bits - NODEBITS)) & NODEMASK);
	}
	items = 0;
	while (1) {
		if (slot == end) {
			if (p == path) {
				index = 0;
				break;
			}
			--p;
			slot = p->slot + 1;	/* matches already with current index */
			end = p->end;
			bits += NODEBITS;
		} else if (*slot) {
			if (bits) {
				p->slot = slot;
				p->end = end;
				++p;
				slot = (*slot)->slots;
				end = slot + NODESLOTS;
				bits -= NODEBITS;
			} else {
				if (items >= maxitems)
					break;
				*buf++ = *slot++;
				++items;
				++index;
			}
		} else {
			++slot;
			if (bits < INDEXBITS)
				index += 1ULL << bits;
		}
	}
	*indexp = index;
	return items;
}

/*
 * Print the contents of the radix tree in human readable form. The passed function pointer
 * should print details on the stored item followed by a newline. It is ok to pass NULL. The
 * format string is used for printing a description of what the tree is about.
 */
void rt_dump(struct radix_tree *rt, void (*dumpdata)(void *), const char *fmt, ...)
{
	unsigned int bits;
	unsigned long long index, mask;
	struct node **slot, **end;
	struct slotpath	*p, path[(INDEXBITS + NODEBITS - 1) / NODEBITS];
	va_list ap;

	index = 0;
	p = path;
	slot = &rt->root;
	end = slot + 1;
	bits = rt->maxbits;
	printf("\nrt_dump ");
	va_start(ap, fmt);
	vprintf(fmt, ap);
	printf(": maxindex=%llx\n", rt->maxindex);
	va_end(ap);
	while (1) {
		if (slot == end) {
			if (p == path) {
				index = 0;
				break;
			}
			--p;
			slot = p->slot + 1;	/* matches already with current index */
			end = p->end;
			bits += NODEBITS;
		} else if (*slot) {
			if (bits) {
				mask = rt->maxindex & ~(unsigned long long)NODEMASK << (bits - NODEBITS);
				printf("%016llx mask=%llx count=%u\n", index, mask, (*slot)->count);
				p->slot = slot;
				p->end = end;
				++p;
				slot = (*slot)->slots;
				end = slot + NODESLOTS;
				bits -= NODEBITS;
			} else {
				printf("%016llx\t%s", index, dumpdata ? " " : "\n");
				if (dumpdata)
					dumpdata(*slot);
				++slot;
				++index;
			}
		} else {
			++slot;
			if (bits < INDEXBITS)
				index += 1ULL << bits;
		}
	}
}
